#include <dns_sd.h>
#include <nameser.h>
#include <sys/socket.h>
#include <net/if.h>
#include <assert.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <pcap.h>
#include <pthread.h>
#include <netdb.h>
#include "dumb_server.h"
#include "simclist.h"   /* use the SimCList library ( http://mij.oltrelinux.com/devel/simclist/ )*/
#include <CoreFoundation/CoreFoundation.h>
#include <sys/time.h>
#include <stdio.h>
#include <netinet/tcp.h>

#define FILE_RECORDING_ENABLE 1
#define FILE_RECORDING_ENABLE2 0
#define DEBUG 2


//Global variables

//PCAP session handle (used by pcap Rx and Tx threads)
pcap_t *pcap_handle;

//Network device for PCAP (en0 = Ethernet; en5 = USB Ethernet)
char* pcap_device;

//Number of WARP nodes
unsigned int numWARPnodes;

//ID of local WARP node
unsigned char localWARPnodeID;

//WARP node control structs
warpnetControl* controlArray[MAX_NUMWARPNODES];

//WARP node traffic structs
warpnetTraffic* trafficArray[MAX_NUMWARPNODES];

//WARP node statistic structs (one per node-to-node link)
warpnetObserve* observeArray[MAX_NUMWARPNODES][MAX_NUMWARPNODES];

//Server status
warpServer_status server_status; 

//Global sequence number incrememnted every time a packet is sent to any node
unsigned char ethPkt_seqNumTx, ethPkt_seqNumRx;

//Time structs for scheduled node updates
struct timeval nextConfirmedNodeUpdate;
struct timeval currentTime;
struct timeval confirmedNodeUpdateTimeout;
struct timezone theTimezone;

//Flag used to initiate confirmed control-ACK handshake
unsigned char doControlHandshakeSoon;

//Linked list of active controllers
// Uses the excellent linked-list library from http://mij.oltrelinux.com/devel/simclist/
list_t controllers_list;
list_t controllers_type_list;

//Mutexes for various processes multiple threads need to perform
pthread_mutex_t	mutex_controlStructs;
pthread_mutex_t	mutex_observeStructs;
pthread_mutex_t	mutex_trafficStructs;
pthread_mutex_t	mutex_updatingControllers;
pthread_mutex_t	mutex_serverStatus;
pthread_mutex_t mutex_pcapTx;
pthread_mutex_t mutex_nodeTimeout;

//For controling this program
pthread_mutex_t commmands;

warpnetControl myControlStruct;
warpnetObserve myObserverStruct;
warpnetTraffic myTrafficStruct;
warpnodeCommand myCommandStruct;


//Current Packet Received;
int currentpacket[MAX_NUMWARPNODES];

//File Path for storing data
char path[100]="/Users/cmcresearcher/Desktop/PacketData.m";

// Permit File Writing
int writing_permit=0;

//Global seqNum
unsigned int seqNum=0;

char *convertstruct2str(void* theStruct){
	unsigned char structType;
	static char string[250]="";
	char tempstr[250]="";
	
	//Extract the first byte of the struct and interpret it as its type
	structType = *(unsigned char*)theStruct;
	
	//Possible Struct Types
	warpnetControl* myControlStruct_ptr;
    warpnetObserve* myObserverStruct_ptr;
	warpnetTraffic* myTrafficStruct_ptr;
	switch(structType)
	{
		case STRUCTID_CONTROL:
			//printf("It is a warpnetControlStruct\n");
			myControlStruct_ptr = (warpnetControl *)theStruct;
			
			sprintf(tempstr,"%i ", myControlStruct_ptr->structType); 
			strcpy(string,tempstr);
			sprintf(tempstr,"%i ", myControlStruct_ptr->nodeID);
		    strcat(string, tempstr);
			sprintf(tempstr,"%i ", myControlStruct_ptr->modOrder);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myControlStruct_ptr->txPower);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myControlStruct_ptr->coding);
			strcat(string, tempstr);
			sprintf(tempstr,"%i \n", myControlStruct_ptr->channel);
			strcat(string, tempstr);
			
			//printf("%s\n", string);
			return string;
			break;
			
		case STRUCTID_OBSERVE:
			//printf("It is a warpnetObserveStruct\n");	
			myObserverStruct_ptr = (warpnetObserve *)theStruct;
			
			sprintf(tempstr,"%i ", myObserverStruct_ptr->structType); 
			strcpy(string,tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->nodeID);
		    strcat(string, tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->partnerID);
			strcat(string, tempstr);
			
			//time eval
			sprintf(tempstr,"%f ", (myObserverStruct_ptr->lastUpdate).tv_sec);
			strcat(string, tempstr);
			sprintf(tempstr,"%f ", (myObserverStruct_ptr->lastUpdate).tv_usec);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->goodPackets);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->partnerBadPackets);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->rxBytes);
			strcat(string, tempstr);	
			sprintf(tempstr,"%i ", myObserverStruct_ptr->txBytes);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myObserverStruct_ptr->otherBadPackets);
			strcat(string, tempstr);
			sprintf(tempstr,"%f \n", myObserverStruct_ptr->throughput);
			strcat(string, tempstr);
			
			//printf("%s\n", string);
			return string;
			break;
			
		case STRUCTID_NODECOMMAND:
			//			typedef struct {
			//				unsigned char structType;
			//				unsigned char nodeID;
			//				unsigned short cmdID;
			//				unsigned int cmdParam;
			//			} warpnodeCommand;
			break;
		case STRUCTID_TRAFFIC:
			myTrafficStruct_ptr = (warpnetTraffic *)theStruct;
			
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->structType); 
			strcpy(string,tempstr);
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->nodeID);
		    strcat(string, tempstr);
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->trafficMode);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->reserved0);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->txInterval);
			strcat(string, tempstr);
			sprintf(tempstr,"%i ", myTrafficStruct_ptr->txPacketLen);
			strcat(string, tempstr);
			sprintf(tempstr,"%f \n", myTrafficStruct_ptr->reserved1);
			strcat(string, tempstr);
			
			return string;
			break;			
			
		default:
			printf("UpdateController: Got unknown struct type (%x)\n", structType);
			//Invalid struct type; quit without sending anyting
			return NULL;
			break;
	}
	
	return NULL;
}


void* convertstr2struct(void* theString, int sockfd){
	unsigned int structType;
	//convert head of string to character
	//in the1cases, separate a fixed number of inputs
	char tempstr[250]="";
	
	//Getting the struct type from the string
	sscanf(theString, "%d%s", &structType, tempstr);
	//printf("\nThe Struct type is: %d",structType);
	
	unsigned int a[10];
	float b;
	
	switch(structType)
	{
		case STRUCTID_CONTROL:
			sscanf(theString, "%i %i %i %i %i %i %s", &a[0], &a[1], &a[2], &a[3], &a[4], &a[5], tempstr);
			myControlStruct.structType=(unsigned char) a[0];
			myControlStruct.nodeID=(unsigned char) a[1];
			myControlStruct.modOrder=(unsigned char) a[2];
			myControlStruct.txPower=(unsigned char) a[3];
			myControlStruct.coding=(unsigned char) a[4];
			myControlStruct.channel=(unsigned char) a[5];
			
			printf("\nThe Control structure is: %i %i %i %i %i %i\n",  myControlStruct.structType, myControlStruct.nodeID,
				   myControlStruct.modOrder, myControlStruct.txPower, myControlStruct.coding, myControlStruct.channel);
			ProcessControllerMessage((unsigned char *) &myControlStruct, 6, sockfd);		//6, struct is 6 bytes long
			break;
			
		case STRUCTID_OBSERVE:
			sscanf(theString, "%i %i %i %i %i %i %i %i %i %i %f %s",  &a[0], &a[1], &a[2], &a[3], &a[4], &a[5], &a[6], &a[7], &a[8], &a[9], &b, tempstr);
			
			myObserverStruct.structType=(unsigned char) a[0];
			myObserverStruct.nodeID=(unsigned char) a[1];
			myObserverStruct.partnerID=(unsigned char) a[2];
			myObserverStruct.lastUpdate.tv_sec=a[3];
			myObserverStruct.lastUpdate.tv_usec=a[4];
			myObserverStruct.goodPackets= a[5];
			myObserverStruct.partnerBadPackets= a[6];
			myObserverStruct.rxBytes= a[7];
			myObserverStruct.txBytes= a[8];
			myObserverStruct.otherBadPackets= a[9];
			myObserverStruct.throughput= b;
			
			printf("\nThe Observe  structure is:\n%i %i %i %i %i %i %i %i %i %i %f\n", myObserverStruct.structType, myObserverStruct.nodeID,
				   myObserverStruct.partnerID, myObserverStruct.lastUpdate.tv_sec, myObserverStruct.lastUpdate.tv_usec, myObserverStruct.goodPackets,
				   myObserverStruct.partnerBadPackets, myObserverStruct.rxBytes, myObserverStruct.txBytes, myObserverStruct.otherBadPackets,
				   myObserverStruct.throughput);
			ProcessControllerMessage((unsigned char *) &myObserverStruct, 36, sockfd);		//40 is to satisfy the largest sruct we have
			break;
			
		case STRUCTID_COMMAND:
			sscanf(theString, "%i %i %i %i %s",  &a[0], &a[1], &a[2], &a[3], tempstr);
			myCommandStruct.structType=(unsigned char) a[0];
			myCommandStruct.nodeID=(unsigned char) a[1];
			myCommandStruct.cmdID=(unsigned short) a[2];
			myCommandStruct.cmdParam=(unsigned int) a[3];
			
			printf("\nThe Command structure is:\n%i %i %i %i\n", myCommandStruct.structType, myCommandStruct.nodeID,
				   myCommandStruct.cmdID, myCommandStruct.cmdParam);
			ProcessControllerMessage((unsigned char *) &myCommandStruct, 8, sockfd);		//40 is to satisfy the largest sruct we have
			break;
			
		case STRUCTID_TRAFFIC:
			sscanf(theString, "%i %i %i %i %i %i %i %s",  &a[0], &a[1], &a[2], &a[3], &a[4], &a[5], &a[6], tempstr);
			myTrafficStruct.structType=(unsigned char) a[0];
			myTrafficStruct.nodeID=(unsigned char) a[1];
			myTrafficStruct.trafficMode=(unsigned char) a[2];
			myTrafficStruct.reserved0= (unsigned char) a[3];
			myTrafficStruct.txInterval= (unsigned int) a[4];
			myTrafficStruct.txPacketLen= (unsigned short) a[5];
			myTrafficStruct.reserved1= (unsigned short) a[6];
			
			printf("\nThe Traffic structure is:\n%i %i %i %i %i %i %i\n", myTrafficStruct.structType, myTrafficStruct.nodeID,
				   myTrafficStruct.trafficMode, myTrafficStruct.reserved0, myTrafficStruct.txInterval, myTrafficStruct.txPacketLen,
				   myTrafficStruct.reserved1);
			ProcessControllerMessage((unsigned char *) &myTrafficStruct, 12, sockfd);		//40 is to satisfy the largest sruct we have
			break;
			
		default:
			printf("UpdateController: Got unknown struct type (%x)\n", structType);
			//Invalid struct type; quit without sending anyting
			
			break;
	}
	return NULL;
}

// Takes in an array, size of the array, and name of new matlab function
// Outputs reformated matlab with data
// Each row represents a packet compared against the array
void FileWrite(float x[], int size, char name[], int count){
	if (writing_permit==1){
		FILE* hFile;
		
		//Open File for appending
		hFile=fopen( path, "a");
		if (hFile == NULL)
		{   //Error
			printf("File not found\n");
		}
		else
		{
			int i;
			//Name of Matlab Function
			fprintf(hFile, "%s(%d,:) ",name, count);
			fprintf(hFile, "=[");
			for (i=0; i<size; i++){
				fprintf(hFile, " %f ",x[i]);
			}
			fprintf(hFile, "]; \n");
			printf("Values Successfully Written\n");
			printf("Packet Number %d \n", count);
			fclose(hFile);
		}
	}
}

// Takes in an array, size of the array, and name of new matlab function
// Outputs reformated matlab with data
// Each row represents a packet compared against the array
void StatWrite(float x[], int size, char name[], int count){
	if (writing_permit==1){
		FILE* hFile;
		
		//Open File for appending
		hFile=fopen( path, "w");
		if (hFile == NULL)
		{   //Error
			printf("File not found\n");
		}
		else
		{
			int i;
			//Name of Matlab Function
			fprintf(hFile, " %s(%d,:) ",name, count);
			fprintf(hFile, "=[");
			for (i=0; i<size; i++){
				fprintf(hFile, " %f ",x[i]);
			}
			fprintf(hFile, "];");
			printf("Values Successfully Written\n");
			printf("Packet Number %d \n", count);
			fclose(hFile);
		}
	}
}


void PrintStruct(void * theStruct)
{
	int i;
	unsigned char structType;
	
	structType = *(unsigned char *)theStruct;
	//return; //disbale most printing for now
	switch(structType)
	{
		case STRUCTID_CONTROL:
			printf("Control Struct Contents:\n");
			printf("\t structType: %d\n", ((warpnetControl *)theStruct)->structType);
			printf("\t nodeID: %d\n", ((warpnetControl *)theStruct)->nodeID);
			printf("\t modOrder: %d\n", ((warpnetControl *)theStruct)->modOrder);
			printf("\t txPower: %d\n", ((warpnetControl *)theStruct)->txPower);
			printf("\t coding: %d\n", ((warpnetControl *)theStruct)->coding);
			printf("\t channel: %d\n", ((warpnetControl *)theStruct)->channel);
			break;
			
		case STRUCTID_COMMAND:
			printf("Command Struct Contents:\n");
			printf("\t structType: %d\n", ((warpnetCommand *)theStruct)->structType);
			printf("\t nodeID: %d\n", ((warpnetCommand *)theStruct)->nodeID);
			printf("\t cmdID: %d\n", ((warpnetCommand *)theStruct)->cmdID);
			printf("\t cmdID: %d\n", ((warpnetCommand *)theStruct)->cmdParam);
			break;
			
		case STRUCTID_OBSERVE:
			printf("Observe Struct Contents:\n");
			printf("\t structType: %d\n", ((warpnetObserve *)theStruct)->structType);
			printf("\t nodeID: %d\n", ((warpnetObserve *)theStruct)->nodeID);
			printf("\t partnerID: %d\n", ((warpnetObserve *)theStruct)->partnerID);
			printf("\t rxBytes: %d\n", ((warpnetObserve *)theStruct)->rxBytes);
			printf("\t txBytes: %d\n", ((warpnetObserve *)theStruct)->txBytes);
			printf("\t goodPackets: %d\n", ((warpnetObserve *)theStruct)->goodPackets);
			printf("\t badPackets: %d\n", ((warpnetObserve *)theStruct)->otherBadPackets);
			printf("\t badPackets: %d\n", ((warpnetObserve *)theStruct)->partnerBadPackets);
			printf("\t throughput: %f\n", ((warpnetObserve *)theStruct)->throughput);
			printf("\t lastUpdate: %d.%d\n",  ((warpnetObserve *)theStruct)->lastUpdate.tv_sec, ((warpnetObserve *)theStruct)->lastUpdate.tv_usec);
			break;
		case STRUCTID_NODESTATS:
			printf("Node Stats Contents:\n");
			printf("Raw: ");
			for(i=0; i<sizeof(warpnodeStats); i++)
				printf("%02x ", *(unsigned char*)(theStruct+i));
			printf("\n");
			printf("\t structType: 0x%x\n", ((warpnodeStats *)theStruct)->structType);
			printf("\t nodeID: %d\n", ((warpnodeStats *)theStruct)->nodeID);
			printf("\t partnerID: %d\n", ((warpnodeStats *)theStruct)->partnerID);
			printf("\t reserved0: %d\n", ((warpnodeStats *)theStruct)->reserved0);
			printf("\t goodPackets: %d (0x%04x)\n", ((warpnodeStats *)theStruct)->goodPackets, ((warpnodeStats *)theStruct)->goodPackets);
			printf("\t badPackets: %d (0x%04x)\n", ((warpnodeStats *)theStruct)->otherBadPackets, ((warpnodeStats *)theStruct)->otherBadPackets);
			printf("\t badPackets: %d (0x%04x)\n", ((warpnodeStats *)theStruct)->partnerBadPackets, ((warpnodeStats *)theStruct)->partnerBadPackets);
			printf("\t rxBytes: %d (0x%04x)\n", ((warpnodeStats *)theStruct)->rxBytes, ((warpnodeStats *)theStruct)->rxBytes);
			printf("\t txBytes: %d (0x%04x)\n", ((warpnodeStats *)theStruct)->txBytes, ((warpnodeStats *)theStruct)->txBytes);
			printf("\t Time: %f\n", ((float) (((warpnodeStats *)theStruct)->time) / XPAR_PPC405_CORE_CLOCK_FREQ_HZ));
			printf("\t Throughput: %f MBits\n", ((((float) (((warpnodeStats *)theStruct)->rxBytes)) * 8 ))  /
				   ((float)((float) (((warpnodeStats *)theStruct)->time) / XPAR_PPC405_CORE_CLOCK_FREQ_HZ)));
			
			break;
		case STRUCTID_CHANEST:
			printf("Chan Est Struct Contents:\n");
			printf("\t structType: %d\n", ((warpnodeChanEst *)theStruct)->structType);
			printf("\t nodeID: %d\n", ((warpnodeChanEst *)theStruct)->nodeID);
			printf("\t partnerID: %d\n", ((warpnodeChanEst *)theStruct)->partnerID);
			break;
		default:
			printf("PrintStruct: Unknown structType!\n");
			break;
	}
	
	return;
}

void ProcessWARPnodeMessage(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	
	if(DEBUG > 1) printf("ProcessWARPnodeMessage: Got a WARP node update\n");
	
	int i;
	
	//Packets from the WARP node must have this structure; indicies are bytes:
	// warpnetEthernetPktHeader struct first:
	//	0-5: Source MAC address (ignored)
	//	6-11: Destination MAC address (ignored)
	//	12-13: Ethertype = WARPNET_ETHTYPE_NODE2SVR
	//	14-15: Number of bytes in full packet, including this header
	//	16: Number of WARPNet structs in packet
	//	17: Reserved (ignored for now)
	// 18:... 1 or more WARPNet warpnode structs
	
	int thisStructType, TxSock, rv;
	unsigned char thisNode, thisPartner;
	unsigned int pktOffset;
	unsigned int structNum, numStructs, pktLen;
	
	warpnetEthernetPktHeader* thePkt;
	
	//Set the default offset to the first struct in the received pkt (skip over the WARPNet ethernet header)
	pktOffset = sizeof(warpnetEthernetPktHeader);
	
	float timeDiff;
	unsigned int bytesDiff;
	
	//Intrepret the received packet's header
	thePkt = (warpnetEthernetPktHeader*)packet;
	
	numStructs = thePkt->numStructs;
	pktLen = thePkt->pktLength;
	
	//Pointers to various WARPNet struct types to help interpret the packet payload below
	warpnodeStats* tmpStats;
	warpnodeStats localStats;
	warpnodeID* tmpNodeID;
	warpnodeCommand* tmpCmd;
	warpnodeChanEst* tmpChanEst;
	
	struct timeval tmpTime;
	
	structNum = 0;
	
	while( (pktOffset < pktLen)  && (structNum < numStructs) )
	{
		thisStructType = (int)( (unsigned char)packet[pktOffset]); //First byte of this struct
		//printf("\n\nRECEIVED A MESSAGE FROM NODE with type: %x \n\n", thisStructType);
		
		switch(thisStructType)
		{
			case STRUCTID_NODESTATS:
				
				if(DEBUG > 1) printf("Got a stats packet\n");
				//Point to this node's struct in the packet buffer
				tmpStats = (warpnodeStats*)(packet + pktOffset);
				
				//Create a local copy of the struct with bytes reordered as needed
				localStats.structType = tmpStats->structType;
				localStats.nodeID = tmpStats->nodeID;
				localStats.partnerID = tmpStats->partnerID;
				localStats.reserved0 = ntohs(tmpStats->reserved0);
				localStats.goodPackets = ntohl(tmpStats->goodPackets);
				localStats.otherBadPackets = ntohl(tmpStats->otherBadPackets);
				localStats.partnerBadPackets = ntohl(tmpStats->partnerBadPackets);
				localStats.rxBytes = ntohl(tmpStats->rxBytes);
				localStats.txBytes = ntohl(tmpStats->txBytes);
				localStats.time = ntohll(tmpStats->time);
				
				//Shortcut for nodeID (it gets used as an index below)
				thisNode = localStats.nodeID;
				thisPartner = localStats.partnerID;
				
				PrintStruct((void *)&localStats);
				
				//Calculate how many bytes were received since the last update
				bytesDiff = (localStats.rxBytes) - ((observeArray[thisNode][thisPartner])->rxBytes);
				
				//Calculate how much time has passed since the last update, using the pcap packet's header's timestamp
				timersub(&(header->ts), &((observeArray[thisNode][thisPartner])->lastUpdate), &tmpTime);
				timeDiff = (float)(tmpTime.tv_sec) + (float)( ((float)(tmpTime.tv_usec))/1000000);
				
				pthread_mutex_lock(&mutex_observeStructs);
				//Calculate the latest throughput in kbps (8(bits/Byte) * 1/1024(b/kb) = 128)
				(observeArray[thisNode][thisPartner])->throughput = (float)((((float)bytesDiff) / timeDiff)/128);
				
				//Fill in the other fields in the node's local struct
				(observeArray[thisNode][thisPartner])->goodPackets = localStats.goodPackets;
				(observeArray[thisNode][thisPartner])->otherBadPackets = localStats.otherBadPackets;
				(observeArray[thisNode][thisPartner])->partnerBadPackets = localStats.partnerBadPackets;
				(observeArray[thisNode][thisPartner])->rxBytes = localStats.rxBytes;
				(observeArray[thisNode][thisPartner])->txBytes = localStats.txBytes;
				(observeArray[thisNode][thisPartner])->lastUpdate = header->ts;
				
				
#if FILE_RECORDING_ENABLE
				
				float _PacketData[9];
				_PacketData[0] = (float)((header->ts).tv_sec);
				_PacketData[1] = (float)(((float)((header->ts).tv_usec))/1000000);
				_PacketData[2] = localStats.goodPackets;
				_PacketData[3] = localStats.otherBadPackets;
				_PacketData[4] = localStats.partnerBadPackets;
				_PacketData[5] = localStats.rxBytes;
				_PacketData[6] = localStats.txBytes;
				_PacketData[7] = ((float) localStats.time)/XPAR_PPC405_CORE_CLOCK_FREQ_HZ;
				_PacketData[8] = ((float) localStats.rxBytes)*8/ (((float)localStats.time)/XPAR_PPC405_CORE_CLOCK_FREQ_HZ);
				//((warpnodeStats *)theStruct)->nodeID
				
				if (localStats.nodeID<MAX_NUMWARPNODES) {
					char name[30]="Node";
					char nodeIDstr[3];
					sprintf(nodeIDstr,"%i", localStats.nodeID); 
					printf("Name is : %s and ID is: %d", name, localStats.nodeID);
					strcat(name, nodeIDstr);
					StatWrite(_PacketData, 9, name, currentpacket[localStats.nodeID]);
					//Increases the currentpacket count
					currentpacket[localStats.nodeID]++; 
				}
				
				
				
				char StatBuffer[500];
				FILE* mFile;
				//Open, read matlab file, and write to buffer for tcl client
				mFile=fopen( path, "rb");
				
				if (mFile == NULL)
				{   //Error
					printf("File not found\n");
				}
				else
				{
					
					fgets(StatBuffer, 500, mFile);
					//sscanf(mFile, "%500c", mbuffer);
					fclose(mFile);
					printf("The File content: %s\n", StatBuffer);
					char command[200]="rm ";
					strcat(command, path);
					system(command);
					
					strcat(StatBuffer, " _statAck_ \n");
					
					int sizeStatBuffer;
					for (i=0; i<500; i++){
						if ( StatBuffer[i]==(char) NULL){
							sizeStatBuffer=i;
							sizeStatBuffer;
							break;
						}
					};
					
					int bytesSent;
					//Send the updated ack struct to all active controllers
					pthread_mutex_lock(&mutex_updatingControllers);
					list_iterator_start(&controllers_list);
					
					while(list_iterator_hasnext(&controllers_list))
					{
						TxSock = *(int *)list_iterator_next(&controllers_list);
						
						//rv = UpdateController(TxSock, (void *)(&tmpCmd));
						//		
						printf("TxSocket is: %u\n", TxSock);
						printf("Sent StatBuffer: %s \n\n\n", StatBuffer);
						bytesSent=send(TxSock, StatBuffer, sizeStatBuffer,0);
						
						if(DEBUG > 1) printf("Sent controller %s Message (nodeID = %d, sock = %d) and sent bytes %d\n", StatBuffer, tmpCmd->nodeID, TxSock, sizeStatBuffer);
						
						if(bytesSent != sizeStatBuffer)
						{
							fprintf(stderr, "UpdateController: didn't send enough bytes!\n");
						}
						
						
					}
					list_iterator_stop(&controllers_list);
					if(DEBUG > 1) printf("Sent stat est struct to %d controllers\n", i);
					
					pthread_mutex_unlock(&mutex_updatingControllers);
					
				}
				
				
#endif
				pthread_mutex_unlock(&mutex_observeStructs);
				
				if(bytesDiff > 0)
				{
					//Send the updated struct to all active controllers
					pthread_mutex_lock(&mutex_updatingControllers);
					list_iterator_start(&controllers_list);
					i = 0;
					while(list_iterator_hasnext(&controllers_list))
					{
						TxSock = *(int *)list_iterator_next(&controllers_list);
						rv = UpdateController(TxSock, (void *)(observeArray[thisNode][thisPartner]));
						i++;
					}
					list_iterator_stop(&controllers_list);
					if(DEBUG > 1) printf("Sent observe struct to %d controllers\n", i);
					pthread_mutex_unlock(&mutex_updatingControllers);
				}
				
				pktOffset += sizeof(warpnodeStats);
				break;
				
			case STRUCTID_NODEID:
				tmpNodeID = (warpnodeID *)(packet + pktOffset);
				thisNode = tmpNodeID->nodeID;
				if(thisNode != NODEID_UNKNOWN)
				{
					//This captured packet is an actual response
					if(DEBUG > 1) printf("Got a node ID response; nodeID = %d\n", thisNode);
					localWARPnodeID = thisNode;
				}
				else
				{
					//This captured packet is actually one sent by this server (or otherwise echoed back); ignore it
				}
				pktOffset += sizeof(warpnodeID);
				break;
				
			case STRUCTID_NODECOMMAND:
				tmpCmd = (warpnodeCommand *)(packet + pktOffset);
				if(DEBUG > 1) printf("Got a node command pkt from a node: node=%x, cmd=0x%x, cmdParam=0x%x\n", tmpCmd->nodeID, ntohs(tmpCmd->cmdID), ntohs(tmpCmd->cmdID));
				switch(ntohs(tmpCmd->cmdID))
			{
				case NODECMD_NODEACK:
					ethPkt_seqNumRx = ntohl(tmpCmd->cmdParam);
					
					if(DEBUG > 1) {
						printf("Got a NODE ACK- seqNum = %d\n", ethPkt_seqNumRx);
						
						char ackBuffer[100]=" _ack_ ";
						int bytesSent=0;
						int bytesToSend=99;
						char seqNumstr[100];
						sprintf(seqNumstr,"%d", ethPkt_seqNumRx);
						strcat(ackBuffer, seqNumstr);
						strcat(ackBuffer, " \n");
						
						//Send the updated ack struct to all active controllers
						pthread_mutex_lock(&mutex_updatingControllers);
						list_iterator_start(&controllers_list);
						
						while(list_iterator_hasnext(&controllers_list))
						{
							TxSock = *(int *)list_iterator_next(&controllers_list);
							
							//rv = UpdateController(TxSock, (void *)(&tmpCmd));
							//		
							printf("TxSocket is: %u\n", TxSock);
							printf("Sent AckBuffer: %s \n\n\n", ackBuffer);
							bytesSent=send(TxSock, ackBuffer, bytesToSend,0);
							
							if(DEBUG > 1) printf("Sent controller %s Message (nodeID = %d, sock = %d) and sent bytes %d\n", ackBuffer, tmpCmd->nodeID, TxSock, bytesSent);
							
							if(bytesSent != bytesToSend)
							{
								fprintf(stderr, "UpdateController: didn't send enough bytes!\n");
							}
						}
						list_iterator_stop(&controllers_list);
						if(DEBUG > 1) printf("Sent chan est struct to %d controllers\n", i);
						
						pthread_mutex_unlock(&mutex_updatingControllers);
					}
					break;
					
				default:
					break;
			}
				pktOffset += sizeof(warpnodeCommand);
				break;
				
			case STRUCTID_CHANEST:
				
				tmpChanEst = (warpnodeChanEst *)(packet + pktOffset);
				if(DEBUG > 1) printf("Got a chan est pkt from node %d (partner=%d)\n", tmpChanEst->nodeID, tmpChanEst->partnerID);
				
				pktOffset += sizeof(warpnodeChanEst);
				
				unsigned int tmp_chanEst_32b[64];
				signed short *tmp_chanEst_16b;
				tmp_chanEst_16b = (signed short *)(&tmp_chanEst_32b);
				complexInt chanEstInt_AA[64];
				
				chanEstFlt chanEstStruct_AA;
				chanEstStruct_AA.structType = STRUCTID_CHANEST;
				chanEstStruct_AA.nodeID = 0;//tmpChanEst->nodeID;
				chanEstStruct_AA.partnerID = 1;//tmpChanEst->partnerID;
				chanEstStruct_AA.antennaID = tmpChanEst->antennaID;
				chanEstStruct_AA.agc_RF = tmpChanEst->agc_RF;
				chanEstStruct_AA.agc_BB = tmpChanEst->agc_BB;
				chanEstStruct_AA.pktStatus = tmpChanEst->pktStatus;
				
				for(i=0; i<64; i++)
				{
					//Each channel estimate is a complex number, with 16-bits each for I/Q
					//Each estimate is written as a 32-bit word in hardware and must be byte-swapped here before being used as two signed shorts
					tmp_chanEst_32b[i] = ntohl(*(unsigned int *)(packet + pktOffset + i*sizeof(unsigned int)));
				}
				
				for(i=0; i<64; i++)
				{
					chanEstInt_AA[i].I = tmp_chanEst_16b[2*i+1];
					chanEstInt_AA[i].Q = tmp_chanEst_16b[2*i];
				}
				
				//Fill in the channel coefficients
				float temp1[64];
				float temp2[64];
				for(i=0; i<64; i++)//****want node id, time stamp for later
				{
					chanEstStruct_AA.chanEsts[i].I = (float)( ((float)(chanEstInt_AA[i].I)) / 32768);
					chanEstStruct_AA.chanEsts[i].Q = (float)( ((float)(chanEstInt_AA[i].Q)) / 32768);
					
					temp1[i]=chanEstStruct_AA.chanEsts[i].I ;
					temp2[i]=chanEstStruct_AA.chanEsts[i].Q;
				}
				
#if FILE_RECORDING_ENABLE2
				((warpnodeStats *)theStruct)->nodeID)
				
				//Temporary Arrays to be passed into FileWrite();
				float pkStatus[1];
				pkStatus[0]=chanEstStruct_AA.pktStatus;
				float time_sec[1];
				time_sec[0]=(float)((header->ts).tv_sec);
				float time_usec[1];
				time_usec[0]= (float)(((float)((header->ts).tv_usec))/1000000);
				float cRF[1];
				cRF[0]=chanEstStruct_AA.agc_RF;
				float cBB[1];
				cBB[0]=chanEstStruct_AA.agc_BB;
				
				printf("PkStatus: %f \n", pkStatus[0]);
				printf("Time Sec: %f \n", time_sec[0]);
				printf("Time USec: %f \n", time_usec[0]);
				
				//Data arrays that are written to the file
				FileWrite(pkStatus, 1, "pkStatus");
				FileWrite( time_sec, 1, "Sec");
				FileWrite( time_usec, 1, "USec");
				FileWrite( cRF, 1, "chanRF");
				FileWrite( cBB, 1, "chanBB");
				FileWrite( temp1, 64, "chanEstsI");
				FileWrite( temp2, 64, "chanEstsQ");
				
				//Increases the currentpacket count
				currentpacket++; 
#endif
				//Send the updated struct to all active controllers
				pthread_mutex_lock(&mutex_updatingControllers);
				list_iterator_start(&controllers_list);
				i = 0;
				while(list_iterator_hasnext(&controllers_list))
				{
					TxSock = *(int *)list_iterator_next(&controllers_list);
					rv = UpdateController(TxSock, (void *)(&chanEstStruct_AA));
					i++;
				}
				list_iterator_stop(&controllers_list);
				if(DEBUG > 1) printf("Sent chan est struct to %d controllers\n", i);
				pthread_mutex_unlock(&mutex_updatingControllers);
				
				pktOffset += (CHAN_EST_NUMBYTES);
				
				break;
				
			default:
				//Unknown struct from the WARP node; ignore the rest of the packet
				printf("ProcessWARPnodeMessage: Unknown struct type (%x) received from WARP node - giving up\n", thisStructType);
				return;
				break;
		}
		structNum++;
	}
	
	return;
}

void *pcapLoopThread(void* theArgs)
{
	//Wrapper function; easier way to call pcap_loop with the right args
	pcap_loop(pcap_handle, -1, (pcap_handler)ProcessWARPnodeMessage, NULL);
	
	//pcap_loop blocks forever, so we should never get here
	pthread_exit(NULL);
}


//This code is based on the excellent libpcap examples at http://www.tcpdump.org/pcap.htm
int SetupPCAP()
{
	
	char errbuf[PCAP_ERRBUF_SIZE];	// Error string
	struct bpf_program fp;		// The compiled filter
	char filter_exp[19];
	bpf_u_int32 mask;		// Our netmask
	bpf_u_int32 net;		// Our IP
	pthread_t tid;             /* variable to hold thread ID */
	
	fprintf(stdout, "Attaching to device %s\n", pcap_device);
	
	sprintf((char *)&filter_exp, "ether proto 0x%04x", WARPNET_ETHTYPE_NODE2SVR);
	
	//Lookup properties for the network device
	// pcap_device is a global variable, defined either by default or as a cmd line argument
	if (pcap_lookupnet(pcap_device, &net, &mask, errbuf) == -1)
	{
		fprintf(stderr, "PCAP: Couldn't get netmask for device %s: %s\n", pcap_device, errbuf);
		net = 0;
		mask = 0;
	}
	
	// Open the session in promiscuous mode with a big enough buffer
	pcap_handle = pcap_open_live(pcap_device, 65535, 1, 100, errbuf);
	
	if (pcap_handle == NULL)
	{
		fprintf(stderr, "PCAP: Couldn't open device %s: %s\n", pcap_device, errbuf);
		exit(-2);
	}
	
	// Compile and apply the filter
	if (pcap_compile(pcap_handle, &fp, filter_exp, 0, net) == -1)
	{
		fprintf(stderr, "PCAP: Couldn't parse filter %s: %s\n", filter_exp, pcap_geterr(pcap_handle));
		exit(-2);
	}
	
	if (pcap_setfilter(pcap_handle, &fp) == -1)
	{
		fprintf(stderr, "PCAP: Couldn't install filter %s: %s\n", filter_exp, pcap_geterr(pcap_handle));
		exit(-2);
	}
	
	fprintf(stdout, "Applied filter '%s'\n", filter_exp);
	
	//Use a separate thread to call the pcap_loop; this thread will run forever so this function can return
	fprintf(stdout, "Launching pcap_loop thread\n");
	pthread_create(&tid, NULL, pcapLoopThread, NULL);
	
	return 0;
}

int UpdateController(int sock, void* theStruct)
{
	
	unsigned char structType;
	int bytesToSend = 0;
	int bytesSent;
	char* structTypeTxt;
	unsigned char nodeID;
	
	//Extract the first byte of the struct and interpret it as its type
	structType = *(unsigned char*)theStruct;
	nodeID = *(unsigned char*)(theStruct+sizeof(unsigned char));
	
	switch(structType)
	{
		case STRUCTID_SERVERSTATUS:
			bytesToSend = sizeof(warpServer_status);
			structTypeTxt = "server status";
			break;
			
		case STRUCTID_CONTROL:
			bytesToSend = sizeof(warpnetControl);
			structTypeTxt = "node control";
			
			break;
			
		case STRUCTID_OBSERVE:
			bytesToSend = sizeof(warpnetObserve);
			structTypeTxt = "node stats";
			break;
			
		case STRUCTID_CHANEST:
			bytesToSend = sizeof(chanEstFlt);// + CHAN_EST_NUMBYTES;
			structTypeTxt = "node chan est";
			break;
			
		default:
			printf("UpdateController: Got unknown struct type (%x); not sending anyting!\n", structType);
			//Invalid struct type; quit without sending anyting
			return -1;
			break;
	}
	
	if(bytesToSend > 0)
	{
		//We assume the calling function has a mutex lock, so we'll blindly use the socket
		if(DEBUG > 1) printf("Sending controller a struct: \n");
		PrintStruct(theStruct);
		
		bytesSent = send(sock, theStruct, bytesToSend, 0);
		
		if(DEBUG > 1) printf("Sent controller %s struct (nodeID = %d, sock = %d)\n", structTypeTxt, nodeID, sock);
		
		if(bytesSent != bytesToSend)
		{
			fprintf(stderr, "UpdateController: didn't send enough bytes! Connection should be closed\n");
			return 1;
		}
	}
	
	return 0;
}


int ProcessControllerMessage(unsigned char * RxMsg, int numRxBytes, int controller_sock)
{
	unsigned char structType;
	unsigned char nodeID;
	int TxSock, rv;
	int bytesProcessed, bytesToSend;
	unsigned char* pktBuf;
	
	printf("%d AND address at %d \n\n", RxMsg[0], &RxMsg[0]);
	structType = RxMsg[0];
	bytesProcessed = 0;
	
	warpnetEthernetPktHeader ethPktHdr;
	
	//Fill in the header for the outgoing Ethernet packet
	memset(&(ethPktHdr.dstAddr), 0xFF, 6); //Broadcast dst address
	memset(&(ethPktHdr.srcAddr), 0x09, 6); //Arbitrary src address; ignored throughout
	ethPktHdr.ethType = htons(WARPNET_ETHTYPE_SVR2NODE);
	
	switch(structType)
	{
		case STRUCTID_TRAFFIC:
			if(numRxBytes < sizeof(warpnetTraffic))
			{
				fprintf(stderr, "ProcessControllerMessage: not enough bytes for control struct (Rcvd %d, expected >= %d)\n", numRxBytes, sizeof(warpnetTraffic));
				return -1;
			}
			else
			{
				
				warpnodeTraffic* myLocalStruct;
				
				myLocalStruct = (warpnodeTraffic*)RxMsg;
				//Extract the nodeID from the received struct
				nodeID = myLocalStruct->nodeID;
				
				warpnodeTraffic localTraffic;
				
				//Point to this node's struct in the packet buffer
				//Create a local copy of the struct with bytes reordered as needed
				localTraffic.structType = STRUCTID_NODETRAFFIC;
				localTraffic.nodeID = myLocalStruct->nodeID;
				localTraffic.trafficMode = myLocalStruct->trafficMode;
				localTraffic.reserved0 = myLocalStruct->reserved0;
				localTraffic.txInterval = ntohl(myLocalStruct->txInterval);
				localTraffic.txPacketLen = ntohs(myLocalStruct->txPacketLen);
				localTraffic.reserved1 = ntohs(myLocalStruct->reserved1);
				
				printf("NODEID is %d\n", nodeID);
				
				//Safely overwrite the corresponding local control struct
				pthread_mutex_lock(&mutex_trafficStructs);
				memcpy(trafficArray[nodeID], RxMsg, sizeof(warpnetTraffic));
				printf("Updated traffic struct:\n");
				pthread_mutex_unlock(&mutex_trafficStructs);
				
				
				//Send the updated control struct via Ethernet
				ethPktHdr.numStructs = 1;
				ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeTraffic));
				
				//Create a buffer for the outgoing packet
				bytesToSend = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeTraffic);
				pktBuf = malloc(bytesToSend);
				
				//A seqNum of 0 instructs the node *not* to send an ACK
				//ethPktHdr.seqNum = 1;
				ethPktHdr.seqNum=seqNum;
				
				//Fill in the outgoing packet buffer
				// For now, we assume controllers and nodes use the same control struct
				memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
				memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), &localTraffic, sizeof(warpnodeTraffic));
				
				pthread_mutex_lock(&mutex_pcapTx);
				rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
				if(rv < 0) pcap_perror(pcap_handle, "");
				pthread_mutex_unlock(&mutex_pcapTx);
				printf("Sent unconfirmed traffic struct to node\n");
				
				//Update the timout values to delay the next confirmed node update
				pthread_mutex_lock(&mutex_nodeTimeout);
				gettimeofday(&currentTime, &theTimezone);
				timeradd(&currentTime, &confirmedNodeUpdateTimeout, &nextConfirmedNodeUpdate);
				doControlHandshakeSoon = 1;
				if(DEBUG > 1) printf("Current time: %d.%06d\n", currentTime.tv_sec, currentTime.tv_usec);
				if(DEBUG > 1) printf("Next confirmed upate: %d.%06d\n", nextConfirmedNodeUpdate.tv_sec, nextConfirmedNodeUpdate.tv_usec);
				pthread_mutex_unlock(&mutex_nodeTimeout);
				
				pthread_mutex_lock(&mutex_updatingControllers);
				//Update all the other controllers with the new control struct
				list_iterator_start(&controllers_list);
				while(list_iterator_hasnext(&controllers_list))
				{
					//Lookup the file descriptor for this controller
					TxSock = *(int *)list_iterator_next(&controllers_list);
					
					//Don't echo the control struct back to the source controller
					if(TxSock != controller_sock)
					{
						rv = UpdateController(TxSock, RxMsg);
					}
				}
				list_iterator_stop(&controllers_list);
				pthread_mutex_unlock(&mutex_updatingControllers);
				
				//Free the packet buffer allocated above
				free(pktBuf);
			}
			bytesProcessed = sizeof(warpnetTraffic);
			break;
			
		case STRUCTID_CONTROL:
			if(numRxBytes < sizeof(warpnetControl))
			{
				fprintf(stderr, "ProcessControllerMessage: not enough bytes for control struct (Rcvd %d, expected >= %d)\n", numRxBytes, sizeof(warpnetControl));
				return -1;
			}
			else
			{
				//Extract the nodeID from the received struct
				nodeID = ((warpnetControl*)RxMsg)->nodeID;
				
				
				//Safely overwrite the corresponding local control struct
				pthread_mutex_lock(&mutex_controlStructs);
				memcpy(controlArray[nodeID], RxMsg, sizeof(warpnetControl));
				printf("Updated control struct:\n");
				PrintStruct(controlArray[nodeID]);
				pthread_mutex_unlock(&mutex_controlStructs);
				
				//Send the updated control struct via Ethernet
				ethPktHdr.numStructs = 1;
				ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeControl));
				
				//Create a buffer for the outgoing packet
				bytesToSend = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeControl);
				pktBuf = malloc(bytesToSend);
				
				//A seqNum of 0 instructs the node *not* to send an ACK
				//ethPktHdr.seqNum = 1;
				ethPktHdr.seqNum=seqNum;
				
				printf("THE FIRST SEQ NUM IS: %u", seqNum);
				
				//Fill in the outgoing packet buffer
				// For now, we assume controllers and nodes use the same control struct
				memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
				memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), controlArray[nodeID], sizeof(warpnodeControl));
				((warpnodeControl*)(pktBuf+sizeof(warpnetEthernetPktHeader))) -> structType =  STRUCTID_NODECONTROL; //STRUCTID_NODETRAFFIC; //
				
				pthread_mutex_lock(&mutex_pcapTx);
				rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
				
				warpnetEthernetPktHeader* myEtherPkt=(void *)pktBuf;
				printf("The seqNum is: %u \n\n\n", myEtherPkt->seqNum);
				
				if(rv < 0) pcap_perror(pcap_handle, "");
				pthread_mutex_unlock(&mutex_pcapTx);
				printf("Sent unconfirmed control struct to node\n");
				
				//Update the timout values to delay the next confirmed node update
				pthread_mutex_lock(&mutex_nodeTimeout);
				gettimeofday(&currentTime, &theTimezone);
				timeradd(&currentTime, &confirmedNodeUpdateTimeout, &nextConfirmedNodeUpdate);
				doControlHandshakeSoon = 1;
				printf("Current time: %d.%06d\n", currentTime.tv_sec, currentTime.tv_usec);
				printf("Next confirmed upate: %d.%06d\n", nextConfirmedNodeUpdate.tv_sec, nextConfirmedNodeUpdate.tv_usec);
				pthread_mutex_unlock(&mutex_nodeTimeout);
				
				pthread_mutex_lock(&mutex_updatingControllers);
				//Update all the other controllers with the new control struct
				list_iterator_start(&controllers_list);
				while(list_iterator_hasnext(&controllers_list))
				{
					//Lookup the file descriptor for this controller
					TxSock = *(int *)list_iterator_next(&controllers_list);
					
					//Don't echo the control struct back to the source controller
					if(TxSock != controller_sock)
					{
						rv = UpdateController(TxSock, RxMsg);
					}
				}
				list_iterator_stop(&controllers_list);
				pthread_mutex_unlock(&mutex_updatingControllers);
				
				//Free the packet buffer allocated above
				free(pktBuf);
			}
			bytesProcessed = sizeof(warpnetControl);
			break;
			
			
		case STRUCTID_COMMAND: //controller to server
			if(numRxBytes < sizeof(warpnetCommand))
			{
				fprintf(stderr, "ProcessControllerMessage: not enough bytes for command struct (Rcvd %d, expected >= %d)\n", numRxBytes, sizeof(warpnetCommand));
				return -1;
			}
			else
			{
				if(DEBUG > 1) printf("Got command struct\n");
				PrintStruct( (warpnetCommand*)RxMsg );
				//Extract the nodeID from the received struct
				nodeID = ((warpnetCommand*)RxMsg)->nodeID;
				
				//Send the command down to the node via Ethernet
				ethPktHdr.numStructs = 1;
				ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeCommand));
				
				ethPktHdr.seqNum=seqNum;
				
				printf("The EthPktHdr for command struct is %d\n", ethPktHdr.seqNum);
				
				//Create a buffer for the outgoing packet
				bytesToSend = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeCommand);
				pktBuf = malloc(bytesToSend);
				
				warpnodeCommand theNodeCmd;
				theNodeCmd.structType = STRUCTID_NODECOMMAND;
				theNodeCmd.nodeID = nodeID;
				theNodeCmd.cmdID = htons(((warpnetCommand*)RxMsg)->cmdID);
				theNodeCmd.cmdParam = htonl(((warpnetCommand*)RxMsg)->cmdParam);
				
				//Fill in the outgoing packet buffer
				// For now, we assume controllers and nodes use the same control struct
				memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
				memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), &theNodeCmd, sizeof(warpnodeCommand));
				
				//Change the structID to a node command (controllers send warpnetCommand)
				((warpnodeCommand*)(pktBuf+sizeof(warpnetEthernetPktHeader))) -> structType = STRUCTID_NODECOMMAND;
				
				pthread_mutex_lock(&mutex_pcapTx);
				rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
				if(rv < 0) pcap_perror(pcap_handle, "");
				pthread_mutex_unlock(&mutex_pcapTx);
				
			}
			bytesProcessed = sizeof(warpnetCommand);
			break;
		default:
			fprintf(stderr, "ProcessControllerMessage: Invalid structID of type %i; ignoring rest of message\n", structType);
			return(-1);
			break;
	}
	
	if(bytesProcessed < numRxBytes)
	{
		//Received message buffer had multiple messages; recur until all are handled
		ProcessControllerMessage((unsigned char * )(RxMsg+bytesProcessed*sizeof(char)), (numRxBytes - bytesProcessed), controller_sock);
	}
	
	return 0;
}

void* ControlConnectionHandler(void* args)
{
	int rv, i;
	int numRxBytes;
	int sockfd;
	unsigned char RxBuf[MAX_TCPDATA];
	
	//Interpret the input argument as a socket file descriptor
	sockfd = (int)args;
	
	printf("New controller connected (sock = %d) (Now %d controllers)\n", sockfd, server_status.numControllers+1);
	
	//When a new controller connects:
	// Update the local server status struct
	// Send the controller the stats structs for the local node
	// Send the controller the control structs for the local node
	pthread_mutex_lock(&mutex_serverStatus);
	server_status.numControllers++;
	pthread_mutex_unlock(&mutex_serverStatus);
	
	pthread_mutex_lock(&mutex_updatingControllers);
	rv = UpdateController(sockfd, (void *)controlArray[localWARPnodeID]);
	for(i=0; i<numWARPnodes; i++)
	{
		rv += UpdateController(sockfd, (void *)observeArray[localWARPnodeID][i]);
	}
	pthread_mutex_unlock(&mutex_updatingControllers);
	
	if(rv > 0)
	{
		//Failed to update the controller; give up
		printf("Connection failed! Exiting thread (%d controllers left).\n", server_status.numControllers-1);
		
		pthread_mutex_lock(&mutex_serverStatus);
		list_delete_at(&controllers_list, list_locate(&controllers_list, &sockfd));
		server_status.numControllers--;
		pthread_mutex_unlock(&mutex_serverStatus);
		pthread_exit(0);
	}
	
	while(1)
	{	
		bzero(RxBuf,MAX_TCPDATA);
		numRxBytes = recv(sockfd, RxBuf, MAX_TCPDATA-1, 0);
		if(numRxBytes < 1)
		{
			printf("Connection closed! Exiting thread (%d controllers left).\n", server_status.numControllers-1);
			pthread_mutex_lock(&mutex_serverStatus);
			list_delete_at(&controllers_list, list_locate(&controllers_list, &sockfd));
			server_status.numControllers--;
			pthread_mutex_unlock(&mutex_serverStatus);
			pthread_exit(0);
		}
		
		if(DEBUG > 1) printf("Got %d bytes (sock = %d)!\n", numRxBytes, sockfd);
		
		//if statement to see if tcl or c program. The &RxBuf is a pointer, you can convert first and then pass to process controllerMessage
		unsigned char aa[1]= "1";
		if(strncmp((char *) &RxBuf[0],(char *) &aa[0], 1)==0)
		{
			printf("Converting String to struct %s\n", RxBuf);
			int temp;
			unsigned int num;
			sscanf((char *) RxBuf, "%i %i %200c ", &temp ,&num, RxBuf);
			seqNum=num;
			printf("SeqNum is %d\n", seqNum); //This updates seqnum
			
			convertstr2struct(RxBuf, sockfd);
		}
		else
		{
			printf("Else: %s", &aa[0]);
			ProcessControllerMessage((unsigned char *)&RxBuf, numRxBytes, sockfd);
		}
		
		//Reset the Rx bytes counter before starting the next receive
		numRxBytes = 0;
	}
	
	return 0;
}

int SetupControlServer(void)
{
	struct   protoent  *ptrp;     /* pointer to a protocol table entry */
	struct   sockaddr_in sad;     /* structure to hold server's address */
	struct   sockaddr_in cad;     /* structure to hold client's address */
	int      sd, sd2;             /* socket descriptors */
	int      port;                /* protocol port number */
	unsigned int      alen;                /* length of address */
	pthread_t  tid;             /* variable to hold thread ID */
	int rv;
	
	int flag = 1;
	int result;	
	alen = sizeof(cad);
	
	memset((char *)&sad,0,sizeof(sad)); /* clear sockaddr structure   */
	sad.sin_family = AF_INET;            /* set family to Internet     */
	sad.sin_addr.s_addr = INADDR_ANY;    /* set the local IP address */
	
	port = SERVICE_PORT;
	sad.sin_port = htons((u_short)port);
	
	//Map TCP transport protocol name to protocol number
	ptrp = getprotobyname("tcp");
	if (ptrp == 0)  {
		fprintf(stderr, "Server: cannot map \"tcp\" to protocol number");
		exit(-3);
	}
	
	//Create a socket
	sd = socket(PF_INET, SOCK_STREAM, ptrp->p_proto);
	if (sd < 0) {
		fprintf(stderr, "Server: socket creation failed\n");
		exit(-3);
	}
	//Disable some buffering (i.e. delay) in the TCP layer for this socket
	// Used code from http://www.unixguide.net/network/socketfaq/2.16.shtml
	//What about IPTOS_LOWDELAY?
	result = setsockopt(sd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(int));
	
	// Bind a local address to the socket
	rv = bind(sd, (struct sockaddr *)&sad, sizeof (sad));
	if (rv < 0) {
		fprintf(stderr,"Server: bind failed\n");
		exit(-3);
	}
	
	// Specify a size of request queue
	rv = listen(sd, MAX_NUMWARPNODES);
	if (rv < 0) {
		fprintf(stderr,"Server: listen failed\n");
		exit(-3);
	}
	
	//Allow access to the sockets that get created below
	// This releases a lock set in main()
	pthread_mutex_unlock(&mutex_updatingControllers);
	
	// Main server loop - accept and handle requests; each connection spawns a thread
	fprintf(stderr, "Server running on port %d\n", SERVICE_PORT);
	while(1)
	{
		//accept() blocks until a new connection comes in
		sd2 = accept(sd, (struct sockaddr *)&cad, &alen); //adds socket to list
		if (sd2 < 0)
		{
			fprintf(stderr, "Server: accept failed\n");
			exit(-3);
		}
		else
		{
			
			//accept() got a valid connection; spawn a thread to handle it
			pthread_mutex_lock(&mutex_serverStatus);
			list_append(&controllers_list, &sd2); //Controllers_list is the list that gets added to with the decision for what path the socket connects to
			
			char Buffer [MAX_TCPDATA]; //THIS IS FOR CONNECTING TO BOARDS
			int numBuffer;
			bzero(Buffer,MAX_TCPDATA); //Set Buffer to zero
			numBuffer = recv(sd, Buffer, MAX_TCPDATA-1, 0);
			printf("\n What was in the Buffer %s \n\n", Buffer);
			
			if(numBuffer < 1)
			{
				printf("Connection closed! Exiting thread (%d controllers left).\n", server_status.numControllers-1);
				list_delete_at(&controllers_list, list_locate(&controllers_list, &sd));
				server_status.numControllers--;
			}
			else
			{
				int type;
				sprintf(Buffer,"%i ", type);
				list_append(&controllers_type_list, &type); //Controllers_list is the list that gets added to with the decision for what path the socket connects to
				printf("\n\n\n THE TYPE IS %d", type);
			}
			
			pthread_mutex_unlock(&mutex_serverStatus);
			pthread_create(&tid, NULL, ControlConnectionHandler, (void *) sd2 );
		}
	}
	close(sd);
	
    return 0;
}

//void IdenfityLocalWARPNode()
//{
//	int rv;
//	unsigned int bytesToSend, numTries;
//	unsigned char* pktBuf;
//	
//	warpnetEthernetPktHeader ethPktHdr;
//	warpnodeID idStruct;
//	
//	//Fill in the header for the outgoing Ethernet packet
//	memset(&(ethPktHdr.dstAddr), 0xFF, 6); //Broadcast dst address
//	memset(&(ethPktHdr.srcAddr), 0x02, 6); //Arbitrary src address; ignored throughout
//	ethPktHdr.ethType = htons(WARPNET_ETHTYPE_SVR2NODE);
//	ethPktHdr.numStructs = 1;
//	ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeID));
//	ethPktHdr.seqNum = 0;
//	
//	//Fill in a "blank" nodeID struct
//	idStruct.structType = STRUCTID_NODEID;
//	idStruct.nodeID = NODEID_UNKNOWN;
//	
//	//Create a buffer for the outgoing packet
//	bytesToSend = sizeof(warpnodeID) + sizeof(warpnetEthernetPktHeader);
//	pktBuf = malloc(MAX_ETHPKTSIZE);
//	
//	//Fill in the outgoing packet buffer
//	memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
//	memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), &(idStruct), sizeof(warpnodeID));
//	
//	//Set default nodeID to invalid value to start
//	localWARPnodeID = 0;//NODEID_UNKNOWN;
//	
//	//This loop keeps running until the local node ID gets set in the thread handling incoming pcap packets
//	numTries = 0;
//	while( (localWARPnodeID == NODEID_UNKNOWN) && (numTries < MAX_NODEID_ATTEMPTS) )
//	{
//		pthread_mutex_lock(&mutex_pcapTx);
//			rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
//			if(rv < 0) pcap_perror(pcap_handle, "");
//		pthread_mutex_unlock(&mutex_pcapTx);
//		
//		usleep(2000000);
//		numTries++;
//		printf("Sent node ID request %d (inject result = %d)\n", numTries, rv);
//	}
//	
//	if(localWARPnodeID == NODEID_UNKNOWN)
//	{
//		//The loop must have excedded the number of allowed attempts; we have to give up
//		printf("Failed to identify local WARP node; exiting!\n");
//		free(pktBuf);
//		exit(-1);
//	}
//	else
//	{
//		//Got a valid WARP node ID
//		// Send the node its latest control struct
//
//		//Fill in the outgoing packet buffer
//		// For now, we assume controllers and nodes use the same control struct
//		//Send the updated control struct via Ethernet
//		bytesToSend = sizeof(warpnodeControl) + sizeof(warpnetEthernetPktHeader);
//		ethPktHdr.numStructs = 1;
//		ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeControl));
//
//		memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
//		memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), controlArray[localWARPnodeID], sizeof(warpnodeControl));
//		((warpnodeControl*)(pktBuf+sizeof(warpnetEthernetPktHeader))) -> structType = STRUCTID_NODECONTROL;
//
//		
//		pthread_mutex_lock(&mutex_pcapTx);
//			rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
//			if(rv < 0) pcap_perror(pcap_handle, "");
//		pthread_mutex_unlock(&mutex_pcapTx);
//	}
//
//	free(pktBuf);
//	
//	return;
//}

void* VerifyNodeControlState()
{
	
	//int rv;
	int bytesToSend;
	unsigned char* pktBuf;
	
	//Generic Ethernet packet header used repeatedly below
	warpnetEthernetPktHeader ethPktHdr;
	
	//Fill in the header for the outgoing Ethernet packet
	memset(&(ethPktHdr.dstAddr), 0xFF, 6); //Broadcast dst address
	memset(&(ethPktHdr.srcAddr), 0x02, 6); //Arbitrary src address; ignored throughout
	ethPktHdr.ethType = htons(WARPNET_ETHTYPE_SVR2NODE);
	ethPktHdr.numStructs = 1;
	ethPktHdr.pktLength = htons(sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeControl));
	
	//Create a buffer for the outgoing packet
	bytesToSend = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeControl);
	pktBuf = malloc(bytesToSend);
	
	if(DEBUG > 1) printf("VerifyNodeControlState: starting timeout loop\n");
	
	while(1)
	{
		if(doControlHandshakeSoon)
		{
			pthread_mutex_lock(&mutex_nodeTimeout);
			gettimeofday(&currentTime, &theTimezone);
			pthread_mutex_unlock(&mutex_nodeTimeout);
			
			if(timercmp(&nextConfirmedNodeUpdate, &currentTime, <))
			{
				
				if(DEBUG > 1) printf("VerifyNodeControlState: Sending confirmed node update\n");
				if(DEBUG > 1) printf("currentTime: %d.%06d :: nextUpdate: %d.%06d\n", currentTime.tv_sec, currentTime.tv_usec, nextConfirmedNodeUpdate.tv_sec, nextConfirmedNodeUpdate.tv_usec);
				//Send the latest control state until the board sends an ACK
				
				//ethPktHdr.seqNum = ++ethPkt_seqNumTx;
				ethPktHdr.seqNum=seqNum;
				
				
				//	while(ethPkt_seqNumTx != ethPkt_seqNumRx)
				//	{
				pthread_mutex_lock(&mutex_controlStructs);
				//Fill in the outgoing packet buffer
				// For now, we assume controllers and nodes use the same control struct
				memcpy(pktBuf, &(ethPktHdr), sizeof(warpnetEthernetPktHeader));
				memcpy(pktBuf+sizeof(warpnetEthernetPktHeader), controlArray[localWARPnodeID], sizeof(warpnodeControl));
				((warpnodeControl*)(pktBuf+sizeof(warpnetEthernetPktHeader))) -> structType = STRUCTID_NODECONTROL;
				pthread_mutex_unlock(&mutex_controlStructs);
				
				//					pthread_mutex_lock(&mutex_pcapTx);
				//					rv = pcap_inject(pcap_handle, (void *)pktBuf, bytesToSend);
				//					if(rv < 0) pcap_perror(pcap_handle, "");
				//					pthread_mutex_unlock(&mutex_pcapTx);
				//					if(DEBUG > 1) printf("Sent control pkt with seqNum %d\n", ethPkt_seqNumTx);
				
				usleep(650000);
				//	}
				
				pthread_mutex_lock(&mutex_nodeTimeout);
				//Lower the flag after a successful handshake
				// The control connection handler will raise this again next time a confirmed exchange is required
				doControlHandshakeSoon = 0;
				pthread_mutex_unlock(&mutex_nodeTimeout);
			}
		}
		else
		{
			sleep(confirmedNodeUpdateTimeout.tv_sec);
			usleep(confirmedNodeUpdateTimeout.tv_usec);
		}
	}
	
	return NULL;
}

// Thread for Interrupting Keyboard. Requires typing the character and then enter
// "e" Begins writing to file
// "s" Stops writing to file
// "r" Resets packet count, clears file
// "d" Deletes current file
void* KeyboardInterrupt(){
	char str[3];
	while (1){
		scanf ("%s",str);
		if (strcmp(str, "e")==0){
			//Enable Writing to File
			writing_permit=1;
			printf("Enable writing to file at: %s \n", path);
		}
		if (strcmp(str, "s")==0){
			//Disable Writing to File
			writing_permit=0;
			printf("Stop writing to file at : %s \n", path);
		}
		if (strcmp(str, "r")==0){
			//Open and Close File to delete contents
			fclose(fopen( path, "w"));
			int i;
			for (i=0; i<MAX_NUMWARPNODES; i++)
			{
				currentpacket[i]=1;
			}
			printf(" Packet Count Reset, File at %s cleared \n", path);
		}
		if (strcmp(str, "d")==0){
			//Delete file
			remove(path);
			printf("File at %s deleted \n", path);
		}
		//		if (strcmp(str, "p")==0){
		//			printf("Path: \n", path);
		//			scanf ("%s",str);
		//			strcpy(path, str);
		//			printf("New Path: \n", path);
		//		}
		//		if (strcmp(str, "h")==0){
		//			printf("Help: Can't help you yet\n", path);
		//		}
	}
	return NULL;
}


int main(int argc, char *argv[])
{
	system("sudo chmod 777 /dev/bpf*");//NEED TO MODIFY THIS for write directory
	//Holds Program Start
	
	/*
	 printf("Data will be written to %s \n", path);
	 printf("Press Enter When Ready to Start the Program...\n");
	 scanf("(space)");
	 printf("Command: \n");
	 */
	
	printf("Starting WARP Stats Server\n");
	
	int i, j;
	pthread_t  tid;
	pthread_t tid2;
	
	//Default number of WARP nodes; can be overridden on the command line
	numWARPnodes = 2;
	
	
	writing_permit=1; //Write to a file
	fclose(fopen( path, "w"));
	
	//Initialize Starting Packet Count
	for (i=0; i<MAX_NUMWARPNODES; i++)
	{
		currentpacket[i]=1;
	}
	
	
	//	if(argc == 1)
	//	{	
	//Set sensible defaults if the user forgets to provide arguments
	pcap_device = "en0";
	printf("No cmd line params; Assuming pcapDevice: %s, numNodes: %d\n", pcap_device, numWARPnodes);
	//}
	//	else if(argc == 2)
	//	{
	//		//User provided an Ethernet interface to use
	//		pcap_device = argv[1];
	//		printf("Cmd line params: pcapDevice: %s\nAssuming numNodes: %d\n", pcap_device, numWARPnodes);
	//	}
	//	else if(argc == 3)
	//	{
	//		//User provided an Ethernet interface to use and number of WARP nodes
	//		pcap_device = argv[1];
	//		numWARPnodes = atoi((char *)argv[2]);
	//		printf("Cmd line params: pcapDevice: %s, numNodes: %d\n", pcap_device, numWARPnodes);
	//	}
	//	else
	//	{
	//		//User didn't provide good arguments; print usage and quit
	//		printf("Usage: %s <Ethernet device> <number of WARP nodes>\nExamples:\n\t%s (defaults to interface en0)\n\t%s en5 3 (use interface en5, 3 WARP nodes)");
	//		exit(-1);
	//	}
	
	//Initialize the global server status struct
	server_status.numWARPnodes = numWARPnodes;
	server_status.numControllers = 0;
	
	//Initialize the global variables
	doControlHandshakeSoon = 0;
	
	//Linked list of active controllers
	// Initialize the list
    list_init(&controllers_list);
	
	// Set the list to store ints, and full copies of values (not just pointers)
	list_attributes_copy(&controllers_list, list_meter_int32_t, 1);
	list_attributes_comparator(&controllers_list, list_comparator_int32_t);
	
	//Initialize the mutex objects used throughout
	pthread_mutex_init(&mutex_controlStructs, NULL);
	pthread_mutex_init(&mutex_observeStructs, NULL);
	pthread_mutex_init(&mutex_trafficStructs, NULL);
	pthread_mutex_init(&mutex_updatingControllers, NULL);
	pthread_mutex_init(&mutex_serverStatus, NULL);
	pthread_mutex_init(&mutex_pcapTx, NULL);
	pthread_mutex_init(&mutex_nodeTimeout, NULL);
	
	//Initialize the timeval structs used to do confirmed handshakes with the local node
	confirmedNodeUpdateTimeout.tv_sec = 0;
	confirmedNodeUpdateTimeout.tv_usec = 500000;
	theTimezone.tz_minuteswest = 0;
	theTimezone.tz_dsttime = DST_USA;
	
	//This is the only thread during initialization, so no need for mutex locks to initialize the global arrays
	for(i=0; i<numWARPnodes; i++)
	{
		//i is the node ID
		//j is each node's partner ID
		
		controlArray[i] = malloc(sizeof(warpnetControl));
		controlArray[i]->structType = STRUCTID_CONTROL;
		controlArray[i]->nodeID = i;
		
		//Choose sensible defaults for the control params
		controlArray[i]->modOrder = 2;	//QPSK
		controlArray[i]->txPower = 63;	//Max power
		controlArray[i]->coding = 0;	//1/2 rate code
		controlArray[i]->channel = 9;	//"Safe" channel
		
		for(j=0; j<numWARPnodes; j++)
		{
			//Initialize the statistics structs
			(observeArray[i][j]) = malloc(sizeof(warpnetObserve));
			(observeArray[i][j])->structType = STRUCTID_OBSERVE;
			(observeArray[i][j])->nodeID = i;
			(observeArray[i][j])->partnerID = j;
			((observeArray[i][j])->lastUpdate).tv_sec = 0;
			((observeArray[i][j])->lastUpdate).tv_usec = 0;
			(observeArray[i][j])->goodPackets = 0;
			(observeArray[i][j])->otherBadPackets = 0;
			(observeArray[i][j])->partnerBadPackets = 0;
			(observeArray[i][j])->rxBytes = 0;
			(observeArray[i][j])->txBytes = 0;
			(observeArray[i][j])->throughput = 0;
		}
		
		//Initialize the traffic structs
		(trafficArray[i]) = malloc(sizeof(warpnetTraffic));
		(trafficArray[i])->structType = STRUCTID_TRAFFIC;
		
	}
	
	//Block any attempts to contact the controllers for now
	// This is cleared after the sockets are setup later in this thread
	pthread_mutex_lock(&mutex_updatingControllers);
	
	
	//Sets up the PCAP capture interface
	// This function returns after creating a thread to handle incoming captures
	SetupPCAP();
	//
	//	//Queries the local WARP node for its wireless network ID
	//	// This function returns only after it successfully identifies the node
	//	// If no node is found, the whole program quits (no point advertising a node-less server)
	//	IdenfityLocalWARPNode();
	//
	//Use a separate thread to send confirmed control packets to the local node
	// This thread runs forever, performing control-ACK handshakes every so often
	pthread_create(&tid, NULL, VerifyNodeControlState, (void *)NULL);
	
	//Keyboard Interrupt Thread
	pthread_create(&tid2, NULL, KeyboardInterrupt, (void *)NULL);
	
	// This function blocks forever, launching a new thread for each connection
	SetupControlServer();
	
	return 0;
}


